import { PolymerElement, html } from '@polymer/polymer/polymer-element.js';
import { Annotation } from './annotation.js';

import './element-editor.js';
import './join-editor.js';
import './split-editor.js';
import './comment-editor.js';

class AnnotationEditor extends PolymerElement {
  static get template() {
    return html`
    <style>
      :host {
        /* set default color for checkbox checked state */
        --paper-checkbox-checked-color: var(--primary-color, var(--paper-blue-500))
      }

      paper-tabs  {
        background: white;
        --paper-tabs-selection-bar-color: var(--paper-blue-900);
      }

      iron-pages section {
        margin: 0 2rem 2rem;
      }
    </style>
    <paper-tabs id="tabs" selected="{{selected}}">
      <paper-tab disabled="[[editDisabled]]">edit</paper-tab>
      <paper-tab disabled="[[splitDisabled]]">split</paper-tab>
      <paper-tab disabled="[[joinDisabled]]">join</paper-tab>
      <paper-tab>comment</paper-tab>
    </paper-tabs>
    <iron-pages selected="[[selected]]">
      <section><element-editor id="elementEditor"></element-editor></section>
      <section><split-editor id="splitEditor"></split-editor></section>
      <section><join-editor id="joinEditor"></join-editor></section>
      <section><comment-editor id="commentEditor"></comment-editor></section>
    </iron-pages>
`;
  }

  static get properties() {
    return {
      editDisabled: {
        type: Boolean,
        value: false
      },
      selectionInfo: {
        type: String,
        value: ''
      },
      elementType: {
        type: String,
        value: 'word'
      },
      selected: {
        value: 0
      }
    }
  }

  connectedCallback() {
    super.connectedCallback();
    this.hidden = true;
    //this.$.tabs.selectIndex(0);

    var editors = ['elementEditor', 'commentEditor', 'joinEditor', 'splitEditor'];
    for (var index = 0; index < editors.length; index++) {
      const editor = this.shadowRoot.getElementById(editors[index]);
      editor.addEventListener('close', event => this.close(event));
    }

  }

  created() {
    this.actions = [];
  }

  setAllowedActions(allowedActions) {
    this.allowedActions = allowedActions;
    this.$.elementEditor.allowedActions = allowedActions;
  }

  setSelection(targets, annotations) {
    this._reset();

    this.$.commentEditor.setTargets(targets);

    // enable editing
    this.editDisabled = (targets.length > 1);
    this.splitDisabled = (targets.length > 1);
    this.joinDisabled = (targets.length === 1);

    function setSelectedTabFromNumberOfTargets(numberOfTargets) {
      if (numberOfTargets === 1) return 0;
      if (numberOfTargets === 3) return 2;
      return 3;
    }
    this.selected = setSelectedTabFromNumberOfTargets(targets.length);

    annotations.forEach(this.setOriginalValue, this);
    this.$.elementEditor.setTargets(targets);
    this.$.joinEditor.setTargets(targets);
    this.$.splitEditor.setTargets(targets);
  }

  loadAnnotations(annotations) {
    console.log('loadAnnotations', annotations);
    annotations.forEach(this.loadAnnotation, this);
  }

  open() {
    this.hidden = false;
    this.$.tabs.notifyResize();
  }

  close() {
    this.hidden = true;
    this._reset();
  }

  loadAnnotation(annotation) {
    console.log('loadAnnotation', annotation);
    if (!annotation) { return }
    switch (annotation.body.getAttribute('subtype')) {
      case Annotation.ActionTypes.JOIN:    // fallthrough
      case Annotation.ActionTypes.SPLIT:   // fallthrough
      case Annotation.ActionTypes.UPDATE:
        this.$.elementEditor.set('text.current', annotation.body.originalValue);
        this.$.elementEditor.set('text.value', annotation.body.display);
        this.$.elementEditor.set('text.annotationId', annotation.id);
        break;
      case Annotation.ActionTypes.REGULARIZE:
        this.$.elementEditor.set('reg.current', annotation.body.originalValue);
        this.$.elementEditor.set('reg.value', annotation.body.display);
        this.$.elementEditor.set('reg.annotationId', annotation.id);
        break;
      case Annotation.ActionTypes.UPDATE_ATTRIBUTE:
        var attribute = annotation.body.querySelector('attribute').getAttribute('name');
        this.$.elementEditor.set([attribute, 'current'], annotation.body.originalValue);
        this.$.elementEditor.set([attribute, 'value'], annotation.body.display);
        this.$.elementEditor.set([attribute, 'annotationId'], annotation.id);
        break;
      case Annotation.ActionTypes.NOTE:
        this.$.elementEditor.set('note.value', annotation.body.display);
        this.$.elementEditor.set('note.annotationId', annotation.id);
        break;
      case Annotation.ActionTypes.COMMENT:
        this.$.commentEditor.set('comment.value', annotation.body.display);
        this.$.commentEditor.set('comment.annotationId', annotation.id);
        this.$.commentEditor.set('private.value', (annotation.visibility === 'private'));
        break;
      default:
        console.log('cannot apply', annotation);
    }
  }

  setOriginalValue(annotation) {
    console.log('setOriginalValue', annotation);
    if (!annotation) { return }
    switch (annotation.body.getAttribute('subtype')) {
      case Annotation.ActionTypes.JOIN:    // fallthrough
      case Annotation.ActionTypes.SPLIT:   // fallthrough
      case Annotation.ActionTypes.UPDATE:
        this.$.elementEditor.set('text.current', annotation.body.originalValue);
        this.$.joinEditor.set('text.current', annotation.body.originalValue);
        this.$.splitEditor.set('text.current', annotation.body.originalValue);
        break;
      case Annotation.ActionTypes.REGULARIZE:
        this.$.elementEditor.set('reg.current', annotation.body.originalValue);
        this.$.joinEditor.set('reg.current', annotation.body.originalValue);
        this.$.splitEditor.set('reg.current', annotation.body.originalValue);
        break;
      case Annotation.ActionTypes.UPDATE_ATTRIBUTE:
        var attribute = annotation.body.querySelector('attribute').getAttribute('name');
        this.$.elementEditor.set([attribute, 'current'], annotation.body.originalValue);
        this.$.joinEditor.set([attribute, 'current'], annotation.body.originalValue);
        this.$.splitEditor.set([attribute, 'current'], annotation.body.originalValue);
        break;
      default:
        console.log('cannot apply', annotation);
    }
  }

  _reset() {
    this.$.elementEditor._reset();
    this.$.commentEditor._reset();
    this.$.joinEditor._reset();
    this.$.splitEditor._reset();
  }
}

customElements.define('annotation-editor', AnnotationEditor);

