import { PolymerElement, html } from '@polymer/polymer/polymer-element.js';
import { IronMeta } from '@polymer/iron-meta/iron-meta.js';

import './annotation-editor.js';
import './selection-info.js';
import './annotation-list.js';

class AnnotationPanel extends PolymerElement {

  static get template() {
    return html`
    <style>
      .page-title {
        @apply(--paper-font-display2);
      }

      .app-menu {
        --paper-menu-selected-item: {
          color: #fff;
          background: var(--light-primary-color);
        }
      }

      paper-menu a > *, paper-menu paper-item > *, paper-menu paper-icon-item > * {
        pointer-events: none;
      }

      @media (max-width: 600px) {
        .page-title {
          font-size: 24px!important;
        }
      }

      h1, h2, h3, h4, h5, h6 {
        font-family: "IM Fell English", serif;
        line-height: 1.2;
      }

      .btn-default:hover {
        color: #333;
        background-color: #e6e6e6;
        border-color: #adadad;
      }

      .btn:not(.btn-raised), .btn:not(.btn-raised).btn-default, .input-group-btn .btn:not(.btn-raised), .input-group-btn .btn:not(.btn-raised).btn-default {
        color: rgba(0, 0, 0, 0.87);
      }

      .btn, .input-group-btn .btn {
        border: none;
        border-radius: 2px;
        margin: 5px 1px;
        text-transform: uppercase;
        letter-spacing: 0;
        will-change: box-shadow, transform;
        transition: box-shadow 0.2s cubic-bezier(0.4, 0, 1, 1), background-color 0.2s cubic-bezier(0.4, 0, 0.2, 1), color 0.2s cubic-bezier(0.4, 0, 0.2, 1);
        outline: 0;
        cursor: pointer;
        text-decoration: none;
        background: 0 0;
      }
    </style>
    <selection-info id="info"></selection-info>
    <annotation-editor id="editor"></annotation-editor>
    <annotation-list server="[[server]]" user="[[user]]" id="list"></annotation-list>
    <slot name="annotations" id="contents" hidden=""></slot>
`;
  }

  static get properties() {
    return {
      server: String,
      user: {
        type: Object,
        value: function () { return {} }
      },
      annotationPromises: {
        type: Number,
        value: 0
      }
    }
  }

  ready() {
    super.ready();
    var meta = new IronMeta;
    this.set('user.name', meta.byKey('user') || 'hieronymus');
    this.set('user.role', meta.byKey('role') || 'reader');
    if (this.user.role !== 'reader') {
      this._editListener = this.$.list.addEventListener('edit', event => this.openEditor(event));
    }

    const editor = this.shadowRoot.getElementById("editor");
    editor.addEventListener('closed', event => this.backToList(event));
    editor.addEventListener('cancelled', event => this.backToList(event));
    editor.addEventListener('saved', event => this.handleAnnotationSave(event));


    // Polymer 3 completely breaks the contract upon which the Polymer 1 app depended in terms
    // of when attached/connected/ready events fire and the ordering of when children elements
    // are loaded in relation to parents.  And <content> is broken and replaced by <slot>.  So
    // we've had to move the slot to the parent (because otherwise it's invisible) and ended up
    // with a mad scramble through the shadow DOM from the parent and a bunch of imperative code
    // to set up initial state.  In the code below, that state consists of the annotations that
    // exist on the page when it is first loaded (as opposed to those the user may add later).

    const slot = this.shadowRoot.getElementById("contents");
    slot.addEventListener('slotchange', e => {
      const slotDiv = slot.assignedNodes()[0];  // should only be one
      var annotations =  slotDiv.querySelectorAll('annotation-item');
      for (var i = 0; i < annotations.length; i++) {
        var a = annotations[i];
        a.apply();
        this.$.list.annotations.push(a);
      }
      this.$.list.close();
    });
  }

  toAtomicList(element) {
    if ('orig' in element) { return element.orig }
    return element
  }
  getTickets(atomicNode) {
    return atomicNode.tickets || []
  }
  flatten(result, tickets) {
    return result.concat(tickets)
  }

  setSelection(annotatableDocument, selection) {
    this.currentDocument = annotatableDocument;
    var tickets = this._getTicketsForSelection(selection);

    // do nothing when a reader clicks on something without a ticket
    if (tickets.length === 0 && this.user.role === 'reader') {
      console.log('tickets found', tickets);
      return this.dispatchEvent(new CustomEvent('cancel.selection', { bubbles: true, composed: true }));
    }
    var acceptedAnnotations = this.$.list.annotations.filter(function (annotation) {
      return (
        tickets.indexOf(annotation.ticket) >= 0 &&
        annotation.status === 'accepted' &&
        'originalValue' in annotation.body
      );
    });
    this.selection = selection;
    this.$.info.setSelection(selection);
    this.$.editor.setSelection(selection, acceptedAnnotations);

    // open editor immediately when no ticket is associated
    if (tickets.length === 0) {
      this.$.list.close();
      this.$.editor.open();
      return;
    }

    this.$.list.filterByTickets(tickets);
    this.$.editor.close();
    this.$.list.open();
  }

  openEditor(e) {
    console.log('openEditor', e);
    if (e.detail && e.detail.data && 'annotations' in e.detail.data) {
      this.$.editor.loadAnnotations(e.detail.data.annotations);
    }
    this.$.list.close();
    this.$.editor.open();
  }

  handleAnnotationSave(event) {
    var myid = this.currentDocument.getAttribute("id");
    var mysource = this.currentDocument.getAttribute("source");
    var myversion = this.currentDocument.getAttribute("version");

    var annotationData = Object.assign({}, event.detail, {
      document: {
        id: myid,
        source: mysource,
        version: myversion
      }
    });
    // edit or reject and add with same ticket number
    // TODO check ticket ID
    if (annotationData.id) {
      return this.$.list.updateAnnotation(annotationData);
    }

    var awaitAddAnnotation = function aaa() {
      this.annotationPromises -= 1;
      if (this.annotationPromises > 0) { return }
      this.backToList();
    }.bind(this);

    // add new annotation
    this.annotationPromises += 1;
    var addPromise = this.$.list.addAnnotation(annotationData);
    addPromise
      .then(awaitAddAnnotation)
      .catch(awaitAddAnnotation);
  }

  reset() {
    this.$.list.close();
    this.$.editor.close();
    this.$.list.resetFilters();
    this.$.info.setSelection();
    this.$.list.removeEventListener('edit', this._editListener);
    this.dispatchEvent(new CustomEvent('cancel.selection', { bubbles: true, composed: true }));
  }

  backToList() {
    this.$.editor.close();
    // wait for new annotations
    if (this.annotationPromises > 0) { return }
    // update list filters
    var tickets = this._getTicketsForSelection(this.selection);
    // check list length
    if (this.$.list.tickets.length === 0 || tickets.length === 0) {
      return this.reset();
    }
    this.$.list.filterByTickets(tickets);
    this.$.list.open();
  }

  _getTicketsForSelection(selection) {
    if (!selection) { return [] }
    return selection.map(this.toAtomicList)
      .map(this.getTickets)
      .reduce(this.flatten, []);
  }

}

customElements.define('annotation-panel', AnnotationPanel);